from ldm2.modules.image_degradation.bsrgan import degradation_bsrgan_variant as degradation_fn_bsr
from ldm2.modules.image_degradation.bsrgan_light import degradation_bsrgan_variant as degradation_fn_bsr_light
